#!/usr/bin/env python

# Edit this script to add your team's code. Some functions are *required*, but you can edit most parts of the required functions,
# change or remove non-required functions, and add your own functions.

################################################################################
#
# Import libraries and functions. You can change or remove them.
#
################################################################################

from helper_code import *
import numpy as np, scipy as sp, scipy.stats, os, sys, joblib
from sklearn.impute import SimpleImputer
from sklearn.ensemble import RandomForestClassifier
from sklearn.utils import resample
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from sklearn.svm import SVC

import opensmile
import pandas as pd

from signal_class import *
from heart_sound_class import *

import warnings
warnings.filterwarnings('ignore')

import tensorflow as tf

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout, Activation
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.callbacks import ModelCheckpoint

from sklearn.preprocessing import LabelEncoder

import soundfile as sf

################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    if verbose >= 1:
        print('Found ', num_patient_files, "patients for training!")

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    murmur_classes = ['Present', 'Unknown', 'Absent']
    classes = ['Present', 'Unknown', 'Absent']
    num_murmur_classes = len(murmur_classes)
    outcome_classes = ['Abnormal', 'Normal']
    num_outcome_classes = len(outcome_classes)

    if verbose >= 1:
        print('Creating dataframe with all the metadata')

    filename_list = []
    patient_id_list =[]
    murmur_list =[]
    outcome_list = []
    feature_list = []

    for i in range(num_patient_files):

        # Load the current patient data
        current_patient_data = load_patient_data(patient_files[i])

        recordings_filenames = extract_recordings_filenames(data_folder, current_patient_data)

        patient_id = get_patient_id(current_patient_data)
        patient_murmur_label = get_murmur(current_patient_data)
        patient_outcome_label = get_outcome(current_patient_data)
        all_murmur_locations = get_all_murmur_locations(current_patient_data)
        all_recording_locations = get_locations(current_patient_data)

        for i, recording in enumerate(recordings_filenames):
            
            filename = os.path.join(recording)

            if patient_murmur_label == "Unknown":
                murmur_sample_label = 2

            elif patient_murmur_label == "Absent":
                murmur_sample_label = 0

            elif patient_murmur_label == "Present":

                location_for_recording = all_recording_locations[i]

                contains_murmur = check_if_audio_location_contains_murmur(location_for_recording, all_murmur_locations)
                
                if contains_murmur:
                    murmur_sample_label = 1

                else:
                    murmur_sample_label = 0

            filename_list.append(filename)
            patient_id_list.append(patient_id)
            murmur_list.append(murmur_sample_label)
            
    metadata_dataframe = pd.DataFrame(list(zip(filename_list, patient_id_list, murmur_list)),
                                            columns = ["filename", "patient_id","murmur"])

    if verbose >= 1:
        print('The metadata was sucessfully created!')

################################################################################

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

   # put the data in a dataframe somehow... which will be called metadata_dataframe

    extract_spectral_features(metadata_dataframe)
    extract_opensmile_features(metadata_dataframe)

    if verbose >= 1:
        print('Features successfully extracted!')

    resampled_metadata_dataframe_for_unknown = balance_training_data_for_unknown(metadata_dataframe, n = 5)

    relabelled_dataframe_for_unknown = relabel_dataframe_for_unknown(resampled_metadata_dataframe_for_unknown)

    # create X_train and y_train
    X_train_for_unknown = relabelled_dataframe_for_unknown.spectral_features.to_list()
    y_train_for_unknown = relabelled_dataframe_for_unknown.murmur.to_list()

    # scale
    scaler_instance_for_unknown = StandardScaler()
    scaler_instance_for_unknown.fit(X_train_for_unknown)

    X_train_for_unknown_scaled = scaler_instance_for_unknown.transform(X_train_for_unknown)

    # reduce
    pca_instance = PCA(0.99)
    pca_instance.fit(X_train_for_unknown_scaled)

    X_train_for_unknown_pcad = pca_instance.transform(X_train_for_unknown_scaled)

    # replace x_train with scaled and pcad feature vectors
    X_train_for_unknown = X_train_for_unknown_pcad

    # instantiate an SVM for unknown sample detection
    clf = SVC(kernel = "linear")

    # time to prepare for murmur training

    resampled_metadata_dataframe_for_murmur = balance_training_data_for_murmur(metadata_dataframe, n = 3)

    X_train_for_murmur = resampled_metadata_dataframe_for_murmur.opensmile_features.to_list()
    y_train_for_murmur = resampled_metadata_dataframe_for_murmur.murmur.to_list()

    # scale
    scaler_instance_for_murmur = StandardScaler()
    scaler_instance_for_murmur.fit(X_train_for_murmur)

    X_train_for_murmur_scaled = scaler_instance_for_murmur.transform(X_train_for_murmur)

    # replace x_train with scaled feature vectors
    X_train_for_murmur = X_train_for_murmur_scaled

    X_train_for_murmur = np.array(X_train_for_murmur)

    labelEncoder = LabelEncoder()
    labelEncoder.fit(y_train_for_murmur)

    encodedLabels_train = labelEncoder.transform(y_train_for_murmur)
    encodedLabelsBinary_train = to_categorical(encodedLabels_train)

    encodedLabelsBinary_train_new = []

    for i, label in enumerate(encodedLabelsBinary_train):

        new_label = np.append(label, 0)
        #print(new_label)

        encodedLabelsBinary_train_new.append(new_label)

    encodedLabelsBinary_train_new = np.array(encodedLabelsBinary_train_new)

    encodedLabelsBinary_train = encodedLabelsBinary_train_new

    y_train_for_murmur = np.array(encodedLabelsBinary_train)

    # create the DNN

    length_of_feature_vector = (X_train_for_murmur[0].shape)[0]

    def build_model_graph(input_shape=(length_of_feature_vector,)):

        model = Sequential()

        model.add(Dense(512))
        model.add(Activation('relu'))
        model.add(Dropout(0.2))

        model.add(Dense(512))
        model.add(Activation('relu'))
        model.add(Dropout(0.5))

        model.add(Dense(256))
        model.add(Activation('relu'))
        model.add(Dropout(0.5))

        model.add(Dense(256))
        model.add(Activation('relu'))
        model.add(Dropout(0.5))

        model.add(Dense(128))
        model.add(Activation('relu'))
        model.add(Dropout(0.5))

        model.add(Dense(128))
        model.add(Activation('relu'))
        model.add(Dropout(0.5))

        model.add(Dense(3))
        model.add(Activation('softmax'))

        # Compile the model
        model.compile(loss='categorical_crossentropy',
                      metrics=['accuracy'], 
                      optimizer='adam')
        return model

    model = build_model_graph(input_shape=(length_of_feature_vector,))

    model.build((None, length_of_feature_vector))
    weights = model.get_weights()

    # Train the model.
    if verbose >= 1:
        print('Training model for unknown sample detection...')

    # train the model for unknown detection
    clf.fit(X_train_for_unknown,y_train_for_unknown)

    if verbose >= 1:
        print('Model for unknown sample detection trained and saved.')

    if verbose >= 1:
        print('Training model for murmur sample detection...')

    # train the DNN for murmur detection
    model.set_weights(weights)

    num_epochs = 30

    checkpoint_filepath = '/tmp/checkpoint'

    callbacks = [EarlyStoppingByLossVal(monitor='loss', value=0.02, verbose=1),
                ModelCheckpoint(checkpoint_filepath, monitor='loss', save_best_only=True, verbose=0)]

    model.fit(X_train_for_murmur, y_train_for_murmur,
          epochs = num_epochs, 
          verbose=verbose, 
          callbacks=callbacks,
          )

    if verbose >= 1:
        print('train accuracy: ', model.evaluate(X_train_for_murmur, y_train_for_murmur)[1])

    # Save the model.
    save_two_model_pipeline(model_folder, classes, scaler_instance_for_unknown, pca_instance, clf, scaler_instance_for_murmur, model)

    if verbose >= 1:
        print('Model for murmur sample detection trained and saved.')

################################################################################

# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):

    filename = os.path.join(model_folder, 'two_model_pipeline.sav')

    return joblib.load(filename)

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def run_challenge_model(model, data, recordings, verbose):

    classes = model['classes']
    scaler_for_unknown = model['scaler_for_unknown']
    pca_for_unknown = model['pca']
    classifier_for_unknown = model["classifier_for_unknown"]
    scaler_for_murmur = model["scaler_for_murmur"]
    classifier_for_murmur = model["classifier_for_murmur"]

    # using data and recordings, create metadata dataframe
    filename_list = []
    patient_id_list = []

    patient_id = get_patient_id(data)
    number_of_recordings = get_num_locations(data)

    for i in range(number_of_recordings):

        new_filename = str(patient_id) + "_" + str(i) + ".wav"

        os.makedirs("/wav_files_for_testing", exist_ok=True)

        new_file_dir = os.path.join("/wav_files_for_testing", new_filename)

        # write audio here
        audio = recordings[i]
        sr = 4000

        sf.write(new_file_dir, audio, sr)

        filename_list.append(new_file_dir)
        patient_id_list.append(patient_id)

    metadata_dataframe = pd.DataFrame(list(zip(filename_list, patient_id_list)),
                                            columns = ["filename", "patient_id"])

    # extract features
    extract_spectral_features(metadata_dataframe)
    extract_opensmile_features(metadata_dataframe)

    # Load features.
    features_for_unknown = metadata_dataframe.spectral_features.to_list()
    features_for_murmur = metadata_dataframe.opensmile_features.to_list()

    # Transform the features for the models
    features_for_unknown_scaled = scaler_for_unknown.transform(features_for_unknown)
    features_for_unknown_pcad = pca_for_unknown.transform(features_for_unknown_scaled)
    features_for_unknown = features_for_unknown_pcad

    features_for_murmur_scaled = scaler_for_murmur.transform(features_for_murmur)
    features_for_murmur = features_for_murmur_scaled
    features_for_murmur = np.array(features_for_murmur)

    # Do the actual classification

    predictions_for_unknown = classifier_for_unknown.predict(features_for_unknown)

    # separate classification for murmur due to uncertainty implementation
     # weight adjustment due to dropout
    mc_weights = classifier_for_murmur.get_weights()
    mc_weights[0] *= 1/0.25     
    classifier_for_murmur.set_weights(mc_weights)

    predictions_stack = np.stack([classifier_for_murmur.predict(features_for_murmur) for sample in range(50)])
    predictions = predictions_stack.mean(axis = 0) # average by column to obtain a more accurate prediction
    predictions_stdev = predictions_stack.std(axis = 0)

    for i, prediction in enumerate(predictions):

        most_likely_label_index = np.argmax(prediction)

        if predictions_stdev[i][most_likely_label_index] > np.mean(predictions_stdev) and most_likely_label_index != 2:

            # create a zeros array the same shape as the prediction
            adjustment_array_of_zeros = np.zeros_like(prediction)

            # create an adjustment array, where the most likely label index
            # gets -0.4 from the probability and unknown gets + 0.4
            current_probability_of_normal = prediction[0]
            current_probability_of_murmur = prediction[1]
            combined_probability = current_probability_of_normal + current_probability_of_murmur

            adjustment_array_for_unknown = adjustment_array_of_zeros
            adjustment_array_for_normal = adjustment_array_of_zeros
            adjustment_array_for_murmur = adjustment_array_of_zeros

            adjustment_array_for_unknown[2] = 0.4
            adjustment_array_for_normal[0] = -0.4 / (combined_probability * current_probability_of_normal)
            adjustment_array_for_murmur[1] = -0.4 / (combined_probability * current_probability_of_murmur)

            adjustment_array = adjustment_array_of_zeros + adjustment_array_for_unknown + adjustment_array_for_normal + adjustment_array_for_murmur
             
            predictions[i] = prediction + adjustment_array
    
    predictions = tf.one_hot(tf.argmax(predictions, axis = 1), depth = 3)
    predictions_for_murmur = predictions.numpy()

    # combine the prediction

    probabilities = []

    for i, _ in enumerate(predictions_for_murmur):

        is_unknown = predictions_for_unknown[i]
        prediction_for_the_given_sample = predictions_for_murmur[i]

        if is_unknown == 1:

            normal_probability = prediction_for_the_given_sample[0]
            murmur_probability = prediction_for_the_given_sample[1]
            combined_probability = normal_probability + murmur_probability

            unknown_probability = 0.45

            updated_normal_probability = (0.55 / combined_probability) * normal_probability
            updated_murmur_probability = (0.55 / combined_probability) * murmur_probability

            updated_prediction = np.array([updated_normal_probability, updated_murmur_probability, unknown_probability])

        else:

            updated_prediction = prediction_for_the_given_sample

        probabilities.append(updated_prediction)

    probabilities_per_sample = probabilities

    list_of_corresponding_patient_ids = metadata_dataframe.patient_id.to_list()

    probabilities_per_patient = {}

    for i, patient in enumerate(list_of_corresponding_patient_ids):

        if patient not in probabilities_per_patient:

            probabilities_per_patient[patient] = list()

        probabilities_per_patient[patient].append(list(probabilities_per_sample[i]))
  
    prediction_mean_list = []
    prediction_std_list = []

    for patient in probabilities_per_patient:

        prediction = probabilities_per_patient[patient]

        prediction_mean = np.mean(prediction, axis=0)

        prediction_std = np.std(prediction, axis=0)

        prediction_mean_list.append(prediction_mean)
        prediction_std_list.append(prediction_std)
    
    # Choose label with higher probability.
    labels = []

    for i, prediction in enumerate(prediction_mean_list):

        prediction_std_sum = prediction[1] + prediction_std_list[i][1]

        if prediction[2] > 0.1:

            label = [0, 0, 1, 1, 0]

        elif prediction_std_sum > 0.3:

            label = [0, 1, 0, 1, 0]

        else:

            label = [1, 0, 0, 0, 1]

        labels.append(label)

    probabilities = prediction_mean_list

    class_order = [1, 2, 0, 3, 4]

    reshaped_probability_list = []
    reshaped_label_list = []

    for probability in probabilities:

        probability = np.concatenate((probability, [probability[1] + probability[2], probability[0]]))
        reshaped_probability = [probability[i] for i in class_order]
        reshaped_probability_list.append(reshaped_probability)

    for label in labels:

        reshaped_label = [label[i] for i in class_order]
        reshaped_label_list.append(reshaped_label)

    probabilities = reshaped_probability_list
    labels = reshaped_label_list

    classes = ['Present', 'Unknown', 'Absent', 'Abnormal', 'Normal']

    return classes, labels, probabilities

################################################################################
#
# Optional functions. You can change or remove these functions and/or add new functions.
#
################################################################################

def extract_recordings_filenames(data_folder, data):

    num_locations = get_num_locations(data)
    recording_information = data.split('\n')[1:num_locations+1]

    recording_filenames = []

    for i in range(num_locations):
        entries = recording_information[i].split(' ')
        recording_file = entries[2]
        filename = os.path.join(data_folder, recording_file)
        recording_filenames.append(filename)

    return recording_filenames

def check_if_audio_location_contains_murmur(location, murmur_locations_for_patient):

    contains_murmur = False
    
    if (murmur_locations_for_patient.find(location) != -1):

        contains_murmur = True

    return contains_murmur

def get_all_murmur_locations(data):

    for l in data.split('\n'):

        if l.startswith('#Murmur locations:'):

            try:
                murmur_location_information = l.split(': ')[1].strip()

            except:
                pass

    return murmur_location_information

# Save your trained model.
def save_two_model_pipeline(model_folder, classes, scaler_for_unknown, pca_instance, svm, scaler_for_murmur, dnn):

    d = {'classes': classes, 
        'scaler_for_unknown': scaler_for_unknown, 
        'pca': pca_instance, 
        'classifier_for_unknown': svm,
        'scaler_for_murmur': scaler_for_murmur,
        "classifier_for_murmur": dnn
        }
    filename = os.path.join(model_folder, 'two_model_pipeline.sav')
    joblib.dump(d, filename, protocol=0)


def extract_spectral_features(metadata_dataframe):

    list_of_filenames = metadata_dataframe.filename.to_list()

    spectral_features_list = []

    for i, sample_filename in enumerate(list_of_filenames):

        sample_audio = HeartSound(sample_filename).audio

        # instantiate Signal

        new_signal_instance = Signal(sample_audio, 4000)

        # on the instance run .extract_spectral_features_from_x_windows()

        spectral_features = new_signal_instance.extract_spectral_features_from_x_windows()

        spectral_features_list.append(spectral_features)

    # flatten the spectral_features

    flat_spectral_features_list = []

    for feature in spectral_features_list:

        flat_feature = [item for sublist in feature for item in sublist]
        flattest_feature = [item for sublist in flat_feature for item in sublist]
        flat_spectral_features_list.append(flattest_feature)

    # save that in the dataframe

    metadata_dataframe["spectral_features"] = flat_spectral_features_list


def extract_opensmile_features(metadata_dataframe):

    smile = opensmile.Smile(
        feature_set=opensmile.FeatureSet.ComParE_2016,
        feature_level=opensmile.FeatureLevel.Functionals,
    )

    opensmile_features_list = []

    # Iterate through each sound file and extract the features 
    for _, row in metadata_dataframe.iterrows():
    
        filename = str(row["filename"])
            
        opensmile_features = smile.process_file(filename)
            
        opensmile_features_array = opensmile_features.to_numpy()
        opensmile_features_array = opensmile_features_array.reshape(6373, )
    
        opensmile_features_list.append(opensmile_features_array)    

    metadata_dataframe["opensmile_features"] = opensmile_features_list


def balance_training_data_for_unknown(metadata_dataframe, n = 3):

    # balance the dataset by resampling

    known_samples = metadata_dataframe[metadata_dataframe["murmur"] != 2]
    unknown_samples  = metadata_dataframe[metadata_dataframe["murmur"] == 2]

    known_samples_downsampled = resample(known_samples, replace=True,
                                    n_samples=len(unknown_samples) * n , random_state = 99)

    unknown_samples_upsampled = resample(unknown_samples, replace = True,
                                    n_samples = len(unknown_samples) * n, random_state = 99)

    resampled_metadata_dataframe = pd.concat([known_samples_downsampled, unknown_samples_upsampled], axis = 0)

    return resampled_metadata_dataframe


def balance_training_data_for_murmur(metadata_dataframe, n = 5):

    murmur_samples = metadata_dataframe[metadata_dataframe["murmur"] == 1]
    normal_samples  = metadata_dataframe[metadata_dataframe["murmur"] == 0]

    normal_samples_downsampled = resample(normal_samples, replace=True,
                                        n_samples=len(murmur_samples) * n , random_state = 99)

    murmur_samples_upsampled = resample(murmur_samples, replace = True,
                                        n_samples = len(murmur_samples) * n, random_state = 99)

    resampled_metadata_dataframe = pd.concat([murmur_samples_upsampled, normal_samples_downsampled], axis = 0)

    return resampled_metadata_dataframe


def relabel_dataframe_for_unknown(metadata_dataframe):

    relabelled_dataframe_for_unknown = metadata_dataframe

    all_labels_from_metadata_dataframe = relabelled_dataframe_for_unknown.murmur.to_list()

    new_label_list = []

    for label in all_labels_from_metadata_dataframe:

        if label == 1 or label == 0:

            new_label = 0

        elif label == 2:

            new_label = 1

        new_label_list.append(new_label)

    relabelled_dataframe_for_unknown["murmur"] = new_label_list

    return relabelled_dataframe_for_unknown


class EarlyStoppingByLossVal(Callback):

    def __init__(self, monitor='loss', value=0.001, verbose=0):
        super(Callback, self).__init__()
        self.monitor = monitor
        self.value = value
        self.verbose = verbose

    def on_epoch_end(self, epoch, logs={}):
        current = logs.get(self.monitor)
        
        if current is None:
            warnings.warn("Early stopping requires %s available!" % self.monitor, RuntimeWarning)

        if current < self.value:
            if self.verbose > 0:
                print("Epoch %05d: early stopping THR" % epoch)
            self.model.stop_training = True


